from collections.abc import Sequence
from types import TracebackType
from typing import final

from asyncpg import (  # pyright: ignore[reportMissingTypeStubs]
    Connection,
    Pool,
    create_pool,  # pyright: ignore [reportUnknownVariableType]
)
from asyncpg.pool import PoolAcquireContext  # pyright: ignore[reportMissingTypeStubs]
from asyncpg.transaction import Transaction  # pyright: ignore[reportMissingTypeStubs]

from integrations.postgres.config import (
    POSTGRES_DATABASE,
    POSTGRES_HOST,
    POSTGRES_PASSWORD,
    POSTGRES_PORT,
    POSTGRES_SSLMODE,
    POSTGRES_USER,
)
from integrations.postgres.state import (
    Postgres,
    PostgresConnection,
    PostgresConnectionContext,
    PostgresTransactionContext,
)
from integrations.postgres.types import (
    PostgresException,
    PostgresRow,
    PostgresValue,
)

__all__ = [
    "PostgresConnectionPool",
]


@final
class PostgresConnectionPool:
    __slots__ = (
        "_connection_limit",
        "_database",
        "_host",
        "_password",
        "_pool",
        "_port",
        "_ssl",
        "_user",
    )

    def __init__(  # noqa: PLR0913
        self,
        host: str = POSTGRES_HOST,
        port: str = POSTGRES_PORT,
        database: str = POSTGRES_DATABASE,
        user: str = POSTGRES_USER,
        password: str = POSTGRES_PASSWORD,
        ssl: str = POSTGRES_SSLMODE,
        connection_limit: int = 1,
    ) -> None:
        self._host: str = host
        self._port: str = port
        self._database: str = database
        self._user: str = user
        self._password: str = password
        self._ssl: str = ssl
        self._connection_limit: int = connection_limit
        self._pool: Pool

    async def __aenter__(self) -> Postgres:
        self._pool = create_pool(
            min_size=1,
            max_size=self._connection_limit,
            database=self._database,
            user=self._user,
            password=self._password,
            host=self._host,
            port=self._port,
            ssl=self._ssl,
        )
        await self._pool  # initialize pool
        return Postgres(prepare_connection=self.prepare_connection)

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        if self._pool._initialized:  # pyright: ignore[reportPrivateUsage]
            await self._pool.close()

    def prepare_connection(self) -> PostgresConnectionContext:
        return _ConnectionContext(pool_context=self._pool.acquire())  # pyright: ignore[reportUnknownMemberType]


@final
class _TransactionContext:
    __slots__ = ("_transaction_context",)

    def __init__(
        self,
        transaction_context: Transaction,
    ) -> None:
        self._transaction_context: Transaction = transaction_context

    async def __aenter__(self) -> None:
        await self._transaction_context.__aenter__()  # pyright: ignore[reportUnknownMemberType]

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        await self._transaction_context.__aexit__(  # pyright: ignore[reportUnknownMemberType]
            exc_type,
            exc_val,
            exc_tb,
        )


@final
class _ConnectionContext:
    __slots__ = ("_pool_context",)

    def __init__(
        self,
        pool_context: PoolAcquireContext,
    ) -> None:
        self._pool_context: PoolAcquireContext = pool_context

    async def __aenter__(self) -> PostgresConnection:
        acquired_connection: Connection = await self._pool_context.__aenter__()  # pyright: ignore[reportUnknownVariableType]

        async def execute(
            statement: str,
            /,
            *args: PostgresValue,
        ) -> Sequence[PostgresRow]:
            try:
                return [
                    dict(record)  # convert to dict to allow match patterns
                    for record in await acquired_connection.fetch(  # pyright: ignore[reportUnknownMemberType, reportUnknownVariableType]
                        statement,
                        *args,
                    )
                ]

            except Exception as exc:
                raise PostgresException("Failed to execute SQL statement") from exc

        def transaction() -> PostgresTransactionContext:
            return _TransactionContext(transaction_context=acquired_connection.transaction())  # pyright: ignore[reportUnknownMemberType, reportUnknownArgumentType]

        return PostgresConnection(
            execute_statement=execute,
            prepare_transaction=transaction,
        )

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        await self._pool_context.__aexit__(  # pyright: ignore[reportUnknownMemberType]
            exc_type,
            exc_val,
            exc_tb,
        )



from haiway.context import (
    Disposable,
    Disposables,
    MetricsContext,
    MetricsHandler,
    MetricsReading,
    MetricsRecording,
    MetricsScopeEntering,
    MetricsScopeExiting,
    MissingContext,
    MissingState,
    ScopeContext,
    ScopeIdentifier,
    StateContext,
    ctx,
)
from haiway.helpers import (
    ArgumentsTrace,
    MetricsHolder,
    MetricsLogger,
    ResultTrace,
    asynchronous,
    cache,
    retry,
    throttle,
    timeout,
    traced,
    wrap_async,
)
from haiway.state import AttributePath, AttributeRequirement, State
from haiway.types import (
    MISSING,
    Default,
    DefaultValue,
    Missing,
    frozenlist,
    is_missing,
    not_missing,
    when_missing,
)
from haiway.utils import (
    AsyncQueue,
    AsyncStream,
    always,
    as_dict,
    as_list,
    as_set,
    as_tuple,
    async_always,
    async_noop,
    freeze,
    getenv_base64,
    getenv_bool,
    getenv_float,
    getenv_int,
    getenv_str,
    load_env,
    mimic_function,
    noop,
    setup_logging,
)

__all__ = [
    "MISSING",
    "ArgumentsTrace",
    "AsyncQueue",
    "AsyncStream",
    "AttributePath",
    "AttributeRequirement",
    "Default",
    "DefaultValue",
    "Disposable",
    "Disposables",
    "MetricsContext",
    "MetricsHandler",
    "MetricsHolder",
    "MetricsLogger",
    "MetricsReading",
    "MetricsRecording",
    "MetricsScopeEntering",
    "MetricsScopeExiting",
    "Missing",
    "MissingContext",
    "MissingState",
    "ResultTrace",
    "ScopeContext",
    "ScopeIdentifier",
    "State",
    "StateContext",
    "always",
    "as_dict",
    "as_list",
    "as_set",
    "as_tuple",
    "async_always",
    "async_noop",
    "asynchronous",
    "cache",
    "ctx",
    "freeze",
    "frozenlist",
    "getenv_base64",
    "getenv_bool",
    "getenv_float",
    "getenv_int",
    "getenv_str",
    "is_missing",
    "load_env",
    "mimic_function",
    "noop",
    "not_missing",
    "retry",
    "setup_logging",
    "throttle",
    "timeout",
    "traced",
    "when_missing",
    "wrap_async",
]



import os
import pkgutil
from collections.abc import Sequence
from importlib import import_module

from haiway import ctx

from integrations.postgres import Postgres, PostgresConnection, PostgresRow
from migrations.postgres.types import Migration

__all__ = [
    "execute_postgres_migrations",
]

# auto load all Migration classes from all files starting with "migration_" and sort by number
MIGRATIONS: Sequence[Migration] = [
    import_module(name=f"migrations.postgres.{name}").execute_migration
    for name, _ in sorted(
        [
            (module, int(module[len("migration_") :]))
            for _, module, _ in pkgutil.iter_modules([os.path.dirname(__file__)])
            if module.startswith("migration_")
        ],
        key=lambda element: int(element[1]),
    )
]


async def execute_postgres_migrations() -> None:
    async with ctx.scope(
        "postgres_migrations",
        disposables=(Postgres.connection(),),
    ):
        # make sure migrations table exists
        await PostgresConnection.execute(MIGRATIONS_TABLE_CREATE_STATEMENT)
        # get current version
        fetched_version: Sequence[PostgresRow] = await PostgresConnection.fetch(
            CURRENT_MIGRATIONS_FETCH_STATEMENT
        )
        current_version: int
        if fetched_version:
            current_version = int(fetched_version[0].get("count", 0))

        else:
            current_version = 0

        ctx.log_info(f"Current database version: {current_version}")

        # perform migrations from current version to latest
        for idx, migration in enumerate(MIGRATIONS[current_version:]):
            ctx.log_info(f"Executing migration {current_version + idx}")
            try:
                async with PostgresConnection.transaction():
                    await migration()

                    await PostgresConnection.execute(MIGRATION_COMPLETION_STATEMENT)

            except Exception as exc:
                ctx.log_info(f"Migration  {current_version + idx} failed")
                raise exc

            else:
                ctx.log_info(f"Migration  {current_version + idx} completed")


MIGRATIONS_TABLE_CREATE_STATEMENT: str = """\
CREATE TABLE IF NOT EXISTS migrations (
    id SERIAL PRIMARY KEY,
    executed_at TIMESTAMP NOT NULL DEFAULT NOW()
);\
"""

CURRENT_MIGRATIONS_FETCH_STATEMENT: str = """\
SELECT COUNT(*) as count FROM migrations;\
"""

# bump version by adding a row to migrations table
MIGRATION_COMPLETION_STATEMENT: str = """\
INSERT INTO migrations DEFAULT VALUES;\
"""



from asyncio import CancelledError, Task, sleep

from pytest import mark, raises

from haiway import timeout


class FakeException(Exception):
    pass


@mark.asyncio
async def test_returns_result_when_returning_value():
    @timeout(3)
    async def long_running() -> int:
        return 42

    assert await long_running() == 42


@mark.asyncio
async def test_raises_with_error():
    @timeout(3)
    async def long_running() -> int:
        raise FakeException()

    with raises(FakeException):
        await long_running()


@mark.asyncio
async def test_raises_with_cancel():
    @timeout(3)
    async def long_running() -> int:
        await sleep(1)
        raise RuntimeError("Invalid state")

    task = Task(long_running())
    with raises(CancelledError):
        await sleep(0.01)
        task.cancel()
        await task


@mark.asyncio
async def test_raises_with_timeout():
    @timeout(0.01)
    async def long_running() -> int:
        await sleep(0.03)
        raise RuntimeError("Invalid state")

    with raises(TimeoutError):
        await long_running()



from contextvars import ContextVar, Token
from types import TracebackType
from typing import Any, Protocol, Self, final, runtime_checkable

from haiway.context.identifier import ScopeIdentifier
from haiway.context.logging import LoggerContext
from haiway.state import State

__all__ = [
    "MetricsContext",
    "MetricsHandler",
    "MetricsReading",
    "MetricsRecording",
    "MetricsScopeEntering",
    "MetricsScopeExiting",
]


@runtime_checkable
class MetricsRecording(Protocol):
    def __call__(
        self,
        scope: ScopeIdentifier,
        /,
        metric: State,
    ) -> None: ...


@runtime_checkable
class MetricsReading(Protocol):
    async def __call__[Metric: State](
        self,
        scope: ScopeIdentifier,
        /,
        *,
        metric: type[Metric],
        merged: bool,
    ) -> Metric | None: ...


@runtime_checkable
class MetricsScopeEntering(Protocol):
    def __call__[Metric: State](
        self,
        scope: ScopeIdentifier,
        /,
    ) -> None: ...


@runtime_checkable
class MetricsScopeExiting(Protocol):
    def __call__[Metric: State](
        self,
        scope: ScopeIdentifier,
        /,
    ) -> None: ...


class MetricsHandler(State):
    record: MetricsRecording
    read: MetricsReading
    enter_scope: MetricsScopeEntering
    exit_scope: MetricsScopeExiting


@final
class MetricsContext:
    _context = ContextVar[Self]("MetricsContext")

    @classmethod
    def scope(
        cls,
        scope: ScopeIdentifier,
        /,
        *,
        metrics: MetricsHandler | None,
    ) -> Self:
        current: Self
        try:  # check for current scope
            current = cls._context.get()

        except LookupError:
            # create root scope when missing
            return cls(
                scope=scope,
                metrics=metrics,
            )

        # create nested scope otherwise
        return cls(
            scope=scope,
            metrics=metrics or current._metrics,
        )

    @classmethod
    def record(
        cls,
        metric: State,
        /,
    ) -> None:
        try:  # catch exceptions - we don't wan't to blow up on metrics
            metrics: Self = cls._context.get()

            if metrics._metrics is not None:
                metrics._metrics.record(
                    metrics._scope,
                    metric,
                )

        except Exception as exc:
            LoggerContext.log_error(
                "Failed to record metric: %s",
                type(metric).__qualname__,
                exception=exc,
            )

    @classmethod
    async def read[Metric: State](
        cls,
        metric: type[Metric],
        /,
        merged: bool,
    ) -> Metric | None:
        try:  # catch exceptions - we don't wan't to blow up on metrics
            metrics: Self = cls._context.get()

            if metrics._metrics is not None:
                return await metrics._metrics.read(
                    metrics._scope,
                    metric=metric,
                    merged=merged,
                )

        except Exception as exc:
            LoggerContext.log_error(
                "Failed to read metric: %s",
                metric.__qualname__,
                exception=exc,
            )

    __slots__ = (
        "_metrics",
        "_scope",
        "_token",
    )

    def __init__(
        self,
        scope: ScopeIdentifier,
        metrics: MetricsHandler | None,
    ) -> None:
        self._scope: ScopeIdentifier
        object.__setattr__(
            self,
            "_scope",
            scope,
        )
        self._metrics: MetricsHandler | None
        object.__setattr__(
            self,
            "_metrics",
            metrics,
        )
        self._token: Token[MetricsContext] | None
        object.__setattr__(
            self,
            "_token",
            None,
        )

    def __setattr__(
        self,
        name: str,
        value: Any,
    ) -> Any:
        raise AttributeError(
            f"Can't modify immutable {self.__class__.__qualname__},"
            f" attribute - '{name}' cannot be modified"
        )

    def __delattr__(
        self,
        name: str,
    ) -> None:
        raise AttributeError(
            f"Can't modify immutable {self.__class__.__qualname__},"
            f" attribute - '{name}' cannot be deleted"
        )

    def __enter__(self) -> None:
        assert self._token is None, "Context reentrance is not allowed"  # nosec: B101
        object.__setattr__(
            self,
            "_token",
            MetricsContext._context.set(self),
        )
        if self._metrics is not None:
            self._metrics.enter_scope(self._scope)

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        assert self._token is not None, "Unbalanced context enter/exit"  # nosec: B101
        MetricsContext._context.reset(self._token)
        object.__setattr__(
            self,
            "_token",
            None,
        )
        if self._metrics is not None:
            self._metrics.exit_scope(self._scope)



from asyncio import (
    CancelledError,
    Task,
    TaskGroup,
    current_task,
    iscoroutinefunction,
)
from collections.abc import (
    AsyncGenerator,
    AsyncIterator,
    Callable,
    Coroutine,
    Iterable,
)
from logging import Logger
from types import TracebackType
from typing import Any, final, overload

from haiway.context.disposables import Disposable, Disposables
from haiway.context.identifier import ScopeIdentifier
from haiway.context.logging import LoggerContext
from haiway.context.metrics import MetricsContext, MetricsHandler
from haiway.context.state import StateContext
from haiway.context.tasks import TaskGroupContext
from haiway.state import State
from haiway.utils import mimic_function
from haiway.utils.stream import AsyncStream

__all__ = [
    "ctx",
]


@final
class ScopeContext:
    __slots__ = (
        "_disposables",
        "_identifier",
        "_logger_context",
        "_metrics_context",
        "_state_context",
        "_task_group_context",
    )

    def __init__(
        self,
        label: str,
        logger: Logger | None,
        task_group: TaskGroup | None,
        state: tuple[State, ...],
        disposables: Disposables | None,
        metrics: MetricsHandler | None,
    ) -> None:
        self._identifier: ScopeIdentifier
        object.__setattr__(
            self,
            "_identifier",
            ScopeIdentifier.scope(label),
        )
        self._logger_context: LoggerContext
        object.__setattr__(
            self,
            "_logger_context",
            LoggerContext(
                self._identifier,
                logger=logger,
            ),
        )
        self._task_group_context: TaskGroupContext | None
        object.__setattr__(
            self,
            "_task_group_context",
            TaskGroupContext(
                task_group=task_group,
            )
            if task_group is not None or self._identifier.is_root
            else None,
        )
        # prepare state context to capture current state
        self._state_context: StateContext
        object.__setattr__(
            self,
            "_state_context",
            StateContext.updated(state),
        )
        self._disposables: Disposables | None
        object.__setattr__(
            self,
            "_disposables",
            disposables,
        )
        self._metrics_context: MetricsContext
        object.__setattr__(
            self,
            "_metrics_context",
            # pre-building metrics context to ensure nested context registering
            MetricsContext.scope(
                self._identifier,
                metrics=metrics,
            ),
        )

    def __setattr__(
        self,
        name: str,
        value: Any,
    ) -> Any:
        raise AttributeError(
            f"Can't modify immutable {self.__class__.__qualname__},"
            f" attribute - '{name}' cannot be modified"
        )

    def __delattr__(
        self,
        name: str,
    ) -> None:
        raise AttributeError(
            f"Can't modify immutable {self.__class__.__qualname__},"
            f" attribute - '{name}' cannot be deleted"
        )

    def __enter__(self) -> str:
        assert (  # nosec: B101
            self._task_group_context is None or self._identifier.is_root
        ), "Can't enter synchronous context with task group"
        assert self._disposables is None, "Can't enter synchronous context with disposables"  # nosec: B101
        self._identifier.__enter__()
        self._logger_context.__enter__()
        self._state_context.__enter__()
        self._metrics_context.__enter__()

        return self._identifier.trace_id

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        self._metrics_context.__exit__(
            exc_type=exc_type,
            exc_val=exc_val,
            exc_tb=exc_tb,
        )

        self._state_context.__exit__(
            exc_type=exc_type,
            exc_val=exc_val,
            exc_tb=exc_tb,
        )

        self._logger_context.__exit__(
            exc_type=exc_type,
            exc_val=exc_val,
            exc_tb=exc_tb,
        )

        self._identifier.__exit__(
            exc_type=exc_type,
            exc_val=exc_val,
            exc_tb=exc_tb,
        )

    async def __aenter__(self) -> str:
        self._identifier.__enter__()
        self._logger_context.__enter__()

        if task_group := self._task_group_context:
            await task_group.__aenter__()

        # lazily initialize state to include disposables results
        if disposables := self._disposables:
            assert self._state_context._token is None  # nosec: B101
            object.__setattr__(
                self,
                "_state_context",
                StateContext(
                    state=self._state_context._state.updated(
                        await disposables.__aenter__(),
                    ),
                ),
            )

        self._state_context.__enter__()
        self._metrics_context.__enter__()

        return self._identifier.trace_id

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        if disposables := self._disposables:
            await disposables.__aexit__(
                exc_type=exc_type,
                exc_val=exc_val,
                exc_tb=exc_tb,
            )

        if task_group := self._task_group_context:
            await task_group.__aexit__(
                exc_type=exc_type,
                exc_val=exc_val,
                exc_tb=exc_tb,
            )

        self._metrics_context.__exit__(
            exc_type=exc_type,
            exc_val=exc_val,
            exc_tb=exc_tb,
        )

        self._state_context.__exit__(
            exc_type=exc_type,
            exc_val=exc_val,
            exc_tb=exc_tb,
        )

        self._logger_context.__exit__(
            exc_type=exc_type,
            exc_val=exc_val,
            exc_tb=exc_tb,
        )

        self._identifier.__exit__(
            exc_type=exc_type,
            exc_val=exc_val,
            exc_tb=exc_tb,
        )

    @overload
    def __call__[Result, **Arguments](
        self,
        function: Callable[Arguments, Coroutine[Any, Any, Result]],
    ) -> Callable[Arguments, Coroutine[Any, Any, Result]]: ...

    @overload
    def __call__[Result, **Arguments](
        self,
        function: Callable[Arguments, Result],
    ) -> Callable[Arguments, Result]: ...

    def __call__[Result, **Arguments](
        self,
        function: Callable[Arguments, Coroutine[Any, Any, Result]] | Callable[Arguments, Result],
    ) -> Callable[Arguments, Coroutine[Any, Any, Result]] | Callable[Arguments, Result]:
        if iscoroutinefunction(function):

            async def async_context(
                *args: Arguments.args,
                **kwargs: Arguments.kwargs,
            ) -> Result:
                async with self:
                    return await function(*args, **kwargs)

            return mimic_function(function, within=async_context)

        else:

            def sync_context(
                *args: Arguments.args,
                **kwargs: Arguments.kwargs,
            ) -> Result:
                with self:
                    return function(*args, **kwargs)  # pyright: ignore[reportReturnType]

            return mimic_function(function, within=sync_context)  # pyright: ignore[reportReturnType]


@final
class ctx:
    __slots__ = ()

    @staticmethod
    def trace_id() -> str:
        """
        Get the current context trace identifier.
        """

        return ScopeIdentifier.current_trace_id()

    @staticmethod
    def scope(
        label: str,
        /,
        *state: State,
        disposables: Disposables | Iterable[Disposable] | None = None,
        logger: Logger | None = None,
        task_group: TaskGroup | None = None,
        metrics: MetricsHandler | None = None,
    ) -> ScopeContext:
        """
        Prepare scope context with given parameters. When called within an existing context\
         it becomes nested with current context as its parent.

        Parameters
        ----------
        label: str
            name of the scope context

        *state: State | Disposable
            state propagated within the scope context, will be merged with current state by\
             replacing current with provided on conflict.

        disposables: Disposables | Iterable[Disposable] | None
            disposables consumed within the context when entered. Produced state will automatically\
             be added to the scope state. Using asynchronous context is required if any disposables\
             were provided.

        logger: Logger | None
            logger used within the scope context, when not provided current logger will be used\
             if any, otherwise the logger with the scope name will be requested.

        task_group: TaskGroup | None
            task group used for spawning and joining tasks within the context. Root scope will
             always have task group created even when not set.

        metrics_store: MetricsStore | None = None
            metrics storage solution responsible for recording and storing metrics.\
             Metrics recroding will be ignored if storage is not provided.
            Assigning metrics_store within existing context will result in an error.

        Returns
        -------
        ScopeContext
            context object intended to enter context manager with.\
             context manager will provide trace_id of current context.
        """

        resolved_disposables: Disposables | None
        match disposables:
            case None:
                resolved_disposables = None

            case Disposables() as disposables:
                resolved_disposables = disposables

            case iterable:
                resolved_disposables = Disposables(*iterable)

        return ScopeContext(
            label=label,
            logger=logger,
            task_group=task_group,
            state=state,
            disposables=resolved_disposables,
            metrics=metrics,
        )

    @staticmethod
    def updated(
        *state: State,
    ) -> StateContext:
        """
        Update scope context with given state. When called within an existing context\
         it becomes nested with current context as its predecessor.

        Parameters
        ----------
        *state: State
            state propagated within the updated scope context, will be merged with current if any\
             by replacing current with provided on conflict

        Returns
        -------
        StateContext
            state part of context object intended to enter context manager with it
        """

        return StateContext.updated(state)

    @staticmethod
    def spawn[Result, **Arguments](
        function: Callable[Arguments, Coroutine[Any, Any, Result]],
        /,
        *args: Arguments.args,
        **kwargs: Arguments.kwargs,
    ) -> Task[Result]:
        """
        Spawn an async task within current scope context task group. When called outside of context\
         it will spawn detached task instead.

        Parameters
        ----------
        function: Callable[Arguments, Coroutine[Any, Any, Result]]
            function to be called within the task group

        *args: Arguments.args
            positional arguments passed to function call

        **kwargs: Arguments.kwargs
            keyword arguments passed to function call

        Returns
        -------
        Task[Result]
            task for tracking function execution and result
        """

        return TaskGroupContext.run(function, *args, **kwargs)

    @staticmethod
    def stream[Element, **Arguments](
        source: Callable[Arguments, AsyncGenerator[Element, None]],
        /,
        *args: Arguments.args,
        **kwargs: Arguments.kwargs,
    ) -> AsyncIterator[Element]:
        """
        Stream results produced by a generator within the proper context state.

        Parameters
        ----------
        source: Callable[Arguments, AsyncGenerator[Result, None]]
            generator streamed as the result

        *args: Arguments.args
            positional arguments passed to generator call

        **kwargs: Arguments.kwargs
            keyword arguments passed to generator call

        Returns
        -------
        AsyncIterator[Result]
            iterator for accessing generated results
        """

        output_stream = AsyncStream[Element]()

        @ctx.scope("stream")
        async def stream() -> None:
            try:
                async for result in source(*args, **kwargs):
                    await output_stream.send(result)

            except BaseException as exc:
                output_stream.finish(exception=exc)

            else:
                output_stream.finish()

        TaskGroupContext.run(stream)
        return output_stream

    @staticmethod
    def check_cancellation() -> None:
        """
        Check if current asyncio task is cancelled, raises CancelledError if so.
        """

        if (task := current_task()) and task.cancelled():
            raise CancelledError()

    @staticmethod
    def cancel() -> None:
        """
        Cancel current asyncio task
        """

        if task := current_task():
            task.cancel()

        else:
            raise RuntimeError("Attempting to cancel context out of asyncio task")

    @staticmethod
    def state[StateType: State](
        state: type[StateType],
        /,
        default: StateType | None = None,
    ) -> StateType:
        """
        Access current scope context state by its type. If there is no matching state defined\
         default value will be created if able, an exception will raise otherwise.

        Parameters
        ----------
        state: type[StateType]
            type of requested state

        Returns
        -------
        StateType
            resolved state instance
        """
        return StateContext.state(
            state,
            default=default,
        )

    @staticmethod
    def record(
        metric: State,
        /,
    ) -> None:
        """
        Record metric within current scope context.

        Parameters
        ----------
        metric: State
            value of metric to be recorded. When a metric implements __add__ it will be added to\
             current value if any, otherwise subsequent calls may replace existing value.

        Returns
        -------
        None
        """

        MetricsContext.record(metric)

    @overload
    @staticmethod
    async def read[Metric: State](
        metric: type[Metric],
        /,
        *,
        merged: bool = False,
    ) -> Metric | None: ...

    @overload
    @staticmethod
    async def read[Metric: State](
        metric: type[Metric],
        /,
        *,
        merged: bool = False,
        default: Metric,
    ) -> Metric: ...

    @staticmethod
    async def read[Metric: State](
        metric: type[Metric],
        /,
        *,
        merged: bool = False,
        default: Metric | None = None,
    ) -> Metric | None:
        """
        Read metric within current scope context.

        Parameters
        ----------
        metric: type[Metric]
            type of metric to be read from current context.

        merged: bool
            control wheather to merge metrics from nested scopes (True)\
             or access only the current scope value (False) without combining them

        default: Metric | None
            default value to return when metric was not recorded yet.

        Returns
        -------
        Metric | None
        """

        value: Metric | None = await MetricsContext.read(
            metric,
            merged=merged,
        )
        if value is None:
            return default

        return value

    @staticmethod
    def log_error(
        message: str,
        /,
        *args: Any,
        exception: BaseException | None = None,
    ) -> None:
        """
        Log using ERROR level within current scope context. When there is no current scope\
         root logger will be used without additional details.

        Parameters
        ----------
        message: str
            message to be written to log

        *args: Any
            message format arguments

        exception: BaseException | None = None
            exception associated with log, when provided full stack trace will be recorded

        Returns
        -------
        None
        """

        LoggerContext.log_error(
            message,
            *args,
            exception=exception,
        )

    @staticmethod
    def log_warning(
        message: str,
        /,
        *args: Any,
        exception: Exception | None = None,
    ) -> None:
        """
        Log using WARNING level within current scope context. When there is no current scope\
         root logger will be used without additional details.

        Parameters
        ----------
        message: str
            message to be written to log

        *args: Any
            message format arguments

        exception: BaseException | None = None
            exception associated with log, when provided full stack trace will be recorded

        Returns
        -------
        None
        """

        LoggerContext.log_warning(
            message,
            *args,
            exception=exception,
        )

    @staticmethod
    def log_info(
        message: str,
        /,
        *args: Any,
    ) -> None:
        """
        Log using INFO level within current scope context. When there is no current scope\
         root logger will be used without additional details.

        Parameters
        ----------
        message: str
            message to be written to log

        *args: Any
            message format arguments

        Returns
        -------
        None
        """

        LoggerContext.log_info(
            message,
            *args,
        )

    @staticmethod
    def log_debug(
        message: str,
        /,
        *args: Any,
        exception: Exception | None = None,
    ) -> None:
        """
        Log using DEBUG level within current scope context. When there is no current scope\
         root logger will be used without additional details.

        Parameters
        ----------
        message: str
            message to be written to log

        *args: Any
            message format arguments

        exception: BaseException | None = None
            exception associated with log, when provided full stack trace will be recorded

        Returns
        -------
        None
        """

        LoggerContext.log_debug(
            message,
            *args,
            exception=exception,
        )



from asyncio import iscoroutinefunction
from collections.abc import Callable, Coroutine, Mapping, Sequence
from typing import Any, Self, cast, overload

from haiway.context import ctx
from haiway.state import State
from haiway.types import MISSING, Missing
from haiway.utils import mimic_function

__all__ = [
    "ArgumentsTrace",
    "ResultTrace",
    "traced",
]


class ArgumentsTrace(State):
    if __debug__:

        @classmethod
        def of(cls, *args: Any, **kwargs: Any) -> Self:
            return cls(
                args=args if args else MISSING,
                kwargs=kwargs if kwargs else MISSING,
            )

    else:  # remove tracing for non debug runs to prevent accidental secret leaks

        @classmethod
        def of(cls, *args: Any, **kwargs: Any) -> Self:
            return cls(
                args=MISSING,
                kwargs=MISSING,
            )

    args: Sequence[Any] | Missing
    kwargs: Mapping[str, Any] | Missing


class ResultTrace(State):
    if __debug__:

        @classmethod
        def of(
            cls,
            value: Any,
            /,
        ) -> Self:
            return cls(result=value)

    else:  # remove tracing for non debug runs to prevent accidental secret leaks

        @classmethod
        def of(
            cls,
            value: Any,
            /,
        ) -> Self:
            return cls(result=MISSING)

    result: Any | Missing


@overload
def traced[**Args, Result](
    function: Callable[Args, Result],
    /,
) -> Callable[Args, Result]: ...


@overload
def traced[**Args, Result](
    *,
    label: str,
) -> Callable[[Callable[Args, Result]], Callable[Args, Result]]: ...


def traced[**Args, Result](
    function: Callable[Args, Result] | None = None,
    /,
    *,
    label: str | None = None,
) -> Callable[[Callable[Args, Result]], Callable[Args, Result]] | Callable[Args, Result]:
    def wrap(
        wrapped: Callable[Args, Result],
    ) -> Callable[Args, Result]:
        if __debug__:
            if iscoroutinefunction(wrapped):
                return cast(
                    Callable[Args, Result],
                    _traced_async(
                        wrapped,
                        label=label or wrapped.__name__,
                    ),
                )

            else:
                return _traced_sync(
                    wrapped,
                    label=label or wrapped.__name__,
                )

        else:  # do not trace on non debug runs
            return wrapped

    if function := function:
        return wrap(wrapped=function)

    else:
        return wrap


def _traced_sync[**Args, Result](
    function: Callable[Args, Result],
    /,
    label: str,
) -> Callable[Args, Result]:
    def traced(
        *args: Args.args,
        **kwargs: Args.kwargs,
    ) -> Result:
        with ctx.scope(label):
            ctx.record(ArgumentsTrace.of(*args, **kwargs))
            try:
                result: Result = function(*args, **kwargs)
                ctx.record(ResultTrace.of(result))
                return result

            except BaseException as exc:
                ctx.record(ResultTrace.of(f"{type(exc)}: {exc}"))
                raise exc

    return mimic_function(
        function,
        within=traced,
    )


def _traced_async[**Args, Result](
    function: Callable[Args, Coroutine[Any, Any, Result]],
    /,
    label: str,
) -> Callable[Args, Coroutine[Any, Any, Result]]:
    async def traced(
        *args: Args.args,
        **kwargs: Args.kwargs,
    ) -> Result:
        with ctx.scope(label):
            ctx.record(ArgumentsTrace.of(*args, **kwargs))
            try:
                result: Result = await function(*args, **kwargs)
                ctx.record(ResultTrace.of(result))
                return result

            except BaseException as exc:
                ctx.record(ResultTrace.of(f"{type(exc)}: {exc}"))
                raise exc

    return mimic_function(
        function,
        within=traced,
    )



from collections.abc import Callable
from typing import Any, cast, final, overload

from haiway.types.missing import MISSING, Missing, not_missing
from haiway.utils.always import always

__all__ = [
    "Default",
    "DefaultValue",
]


@final
class DefaultValue[Value]:
    __slots__ = ("_value",)

    @overload
    def __init__(
        self,
        value: Value,
        /,
    ) -> None: ...

    @overload
    def __init__(
        self,
        /,
        *,
        factory: Callable[[], Value],
    ) -> None: ...

    @overload
    def __init__(
        self,
        value: Value | Missing,
        /,
        *,
        factory: Callable[[], Value] | Missing,
    ) -> None: ...

    def __init__(
        self,
        value: Value | Missing = MISSING,
        /,
        *,
        factory: Callable[[], Value] | Missing = MISSING,
    ) -> None:
        assert (  # nosec: B101
            value is MISSING or factory is MISSING
        ), "Can't specify both default value and factory"

        self._value: Callable[[], Value | Missing]
        if not_missing(factory):
            object.__setattr__(
                self,
                "_value",
                factory,
            )

        else:
            object.__setattr__(
                self,
                "_value",
                always(value),
            )

    def __call__(self) -> Value | Missing:
        return self._value()

    def __setattr__(
        self,
        __name: str,
        __value: Any,
    ) -> None:
        raise AttributeError("Missing can't be modified")

    def __delattr__(
        self,
        __name: str,
    ) -> None:
        raise AttributeError("Missing can't be modified")


@overload
def Default[Value](
    value: Value,
    /,
) -> Value: ...


@overload
def Default[Value](
    *,
    factory: Callable[[], Value],
) -> Value: ...


def Default[Value](
    value: Value | Missing = MISSING,
    /,
    *,
    factory: Callable[[], Value] | Missing = MISSING,
) -> Value:  # it is actually a DefaultValue, but type checker has to be fooled most some cases
    return cast(
        Value,
        DefaultValue(
            value,
            factory=factory,
        ),
    )



from asyncio import gather
from collections.abc import Iterable
from contextlib import AbstractAsyncContextManager
from itertools import chain
from types import TracebackType
from typing import Any, final

from haiway.state import State

__all__ = [
    "Disposable",
    "Disposables",
]

type Disposable = AbstractAsyncContextManager[Iterable[State] | State | None]


@final
class Disposables:
    __slots__ = ("_disposables",)

    def __init__(
        self,
        *disposables: Disposable,
    ) -> None:
        self._disposables: tuple[Disposable, ...]
        object.__setattr__(
            self,
            "_disposables",
            disposables,
        )

    def __setattr__(
        self,
        name: str,
        value: Any,
    ) -> Any:
        raise AttributeError(
            f"Can't modify immutable {self.__class__.__qualname__},"
            f" attribute - '{name}' cannot be modified"
        )

    def __delattr__(
        self,
        name: str,
    ) -> None:
        raise AttributeError(
            f"Can't modify immutable {self.__class__.__qualname__},"
            f" attribute - '{name}' cannot be deleted"
        )

    def __bool__(self) -> bool:
        return len(self._disposables) > 0

    async def _initialize(
        self,
        disposable: Disposable,
        /,
    ) -> Iterable[State]:
        match await disposable.__aenter__():
            case None:
                return ()

            case State() as single:
                return (single,)

            case multiple:
                return multiple

    async def __aenter__(self) -> Iterable[State]:
        return [
            *chain.from_iterable(
                state
                for state in await gather(
                    *[self._initialize(disposable) for disposable in self._disposables],
                )
            )
        ]

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        results: list[bool | BaseException | None] = await gather(
            *[
                disposable.__aexit__(
                    exc_type,
                    exc_val,
                    exc_tb,
                )
                for disposable in self._disposables
            ],
            return_exceptions=True,
        )

        exceptions: list[BaseException] = [exc for exc in results if isinstance(exc, BaseException)]

        if len(exceptions) > 1:
            raise BaseExceptionGroup("Disposing errors", exceptions)



from asyncio import CancelledError, Task, sleep
from time import time
from unittest import TestCase

from pytest import mark, raises

from haiway import retry


class FakeException(Exception):
    pass


@mark.asyncio
async def test_returns_value_without_errors():
    executions: int = 0

    @retry
    def compute(value: str, /) -> str:
        nonlocal executions
        executions += 1
        return value

    assert compute("expected") == "expected"
    assert executions == 1


@mark.asyncio
async def test_retries_with_errors():
    executions: int = 0

    @retry
    def compute(value: str, /) -> str:
        nonlocal executions
        executions += 1
        if executions == 1:
            raise FakeException()
        else:
            return value

    assert compute("expected") == "expected"
    assert executions == 2


@mark.asyncio
async def test_logs_issue_with_errors():
    executions: int = 0
    test_case = TestCase()

    @retry
    def compute(value: str, /) -> str:
        nonlocal executions
        executions += 1
        if executions == 1:
            raise FakeException("fake")
        else:
            return value

    with test_case.assertLogs() as logs:
        compute("expected")
        assert executions == 2
        assert logs.output == [
            f"ERROR:root:Attempting to retry {compute.__name__}"
            f" which failed due to an error: {FakeException('fake')}"
        ]


@mark.asyncio
async def test_fails_with_exceeding_errors():
    executions: int = 0

    @retry(limit=1)
    def compute(value: str, /) -> str:
        nonlocal executions
        executions += 1
        raise FakeException()

    with raises(FakeException):
        compute("expected")
    assert executions == 2


@mark.asyncio
async def test_fails_with_cancellation():
    executions: int = 0

    @retry(limit=1)
    def compute(value: str, /) -> str:
        nonlocal executions
        executions += 1
        raise CancelledError()

    with raises(CancelledError):
        compute("expected")
    assert executions == 1


@mark.asyncio
async def test_retries_with_selected_errors():
    executions: int = 0

    @retry
    def compute(value: str, /) -> str:
        nonlocal executions
        executions += 1
        if executions == 1:
            raise FakeException()
        else:
            return value

    assert compute("expected") == "expected"
    assert executions == 2


@mark.asyncio
async def test_fails_with_not_selected_errors():
    executions: int = 0

    @retry(catching={ValueError})
    def compute(value: str, /) -> str:
        nonlocal executions
        executions += 1
        raise FakeException()

    with raises(FakeException):
        compute("expected")

    assert executions == 1


@mark.asyncio
async def test_async_returns_value_without_errors():
    executions: int = 0

    @retry
    async def compute(value: str, /) -> str:
        nonlocal executions
        executions += 1
        return value

    assert await compute("expected") == "expected"
    assert executions == 1


@mark.asyncio
async def test_async_retries_with_errors():
    executions: int = 0

    @retry
    async def compute(value: str, /) -> str:
        nonlocal executions
        executions += 1
        if executions == 1:
            raise FakeException()
        else:
            return value

    assert await compute("expected") == "expected"
    assert executions == 2


@mark.asyncio
async def test_async_fails_with_exceeding_errors():
    executions: int = 0

    @retry(limit=1)
    async def compute(value: str, /) -> str:
        nonlocal executions
        executions += 1
        raise FakeException()

    with raises(FakeException):
        await compute("expected")
    assert executions == 2


@mark.asyncio
async def test_async_fails_with_cancellation():
    executions: int = 0

    @retry(limit=1)
    async def compute(value: str, /) -> str:
        nonlocal executions
        executions += 1
        raise CancelledError()

    with raises(CancelledError):
        await compute("expected")
    assert executions == 1


@mark.asyncio
async def test_async_fails_when_cancelled():
    executions: int = 0

    @retry(limit=1)
    async def compute(value: str, /) -> str:
        nonlocal executions
        executions += 1
        await sleep(1)
        return value

    with raises(CancelledError):
        task = Task(compute("expected"))
        await sleep(0.02)
        task.cancel()
        await task
    assert executions == 1


@mark.asyncio
async def test_async_uses_delay_with_errors():
    executions: int = 0

    @retry(limit=2, delay=0.05)
    async def compute(value: str, /) -> str:
        nonlocal executions
        executions += 1
        raise FakeException()

    time_start: float = time()
    with raises(FakeException):
        await compute("expected")
    assert (time() - time_start) >= 0.1
    assert executions == 3


@mark.asyncio
async def test_async_uses_computed_delay_with_errors():
    executions: int = 0

    @retry(limit=2, delay=lambda attempt, _: attempt * 0.035)
    async def compute(value: str, /) -> str:
        nonlocal executions
        executions += 1
        raise FakeException()

    time_start: float = time()
    with raises(FakeException):
        await compute("expected")
    assert (time() - time_start) >= 0.1
    assert executions == 3


@mark.asyncio
async def test_async_logs_issue_with_errors():
    executions: int = 0
    test_case = TestCase()

    @retry
    async def compute(value: str, /) -> str:
        nonlocal executions
        executions += 1
        if executions == 1:
            raise FakeException("fake")
        else:
            return value

    with test_case.assertLogs() as logs:
        await compute("expected")
        assert executions == 2
        assert logs.output[0].startswith(
            f"ERROR:root:Attempting to retry {compute.__name__} which failed due to an error"
        )


@mark.asyncio
async def test_async_retries_with_selected_errors():
    executions: int = 0

    @retry
    async def compute(value: str, /) -> str:
        nonlocal executions
        executions += 1
        if executions == 1:
            raise FakeException()
        else:
            return value

    assert await compute("expected") == "expected"
    assert executions == 2


@mark.asyncio
async def test_async_fails_with_not_selected_errors():
    executions: int = 0

    @retry(catching={ValueError})
    async def compute(value: str, /) -> str:
        nonlocal executions
        executions += 1
        raise FakeException()

    with raises(FakeException):
        await compute("expected")

    assert executions == 1



from contextvars import ContextVar, Token
from types import TracebackType
from typing import Any, Self, final
from uuid import uuid4

__all__ = [
    "ScopeIdentifier",
]


@final
class ScopeIdentifier:
    _context = ContextVar[Self]("ScopeIdentifier")

    @classmethod
    def current_trace_id(cls) -> str:
        try:
            return ScopeIdentifier._context.get().trace_id

        except LookupError as exc:
            raise RuntimeError("Attempting to access scope identifier outside of scope") from exc

    @classmethod
    def scope(
        cls,
        label: str,
        /,
    ) -> Self:
        current: Self
        try:  # check for current scope
            current = cls._context.get()

        except LookupError:
            # create root scope when missing
            trace_id: str = uuid4().hex
            return cls(
                label=label,
                scope_id=uuid4().hex,
                parent_id=trace_id,  # trace_id is parent_id for root
                trace_id=trace_id,
            )

        # create nested scope otherwise
        return cls(
            label=label,
            scope_id=uuid4().hex,
            parent_id=current.scope_id,
            trace_id=current.trace_id,
        )

    __slots__ = (
        "_token",
        "label",
        "parent_id",
        "scope_id",
        "trace_id",
        "unique_name",
    )

    def __init__(
        self,
        trace_id: str,
        parent_id: str,
        scope_id: str,
        label: str,
    ) -> None:
        self.trace_id: str
        object.__setattr__(
            self,
            "trace_id",
            trace_id,
        )
        self.parent_id: str
        object.__setattr__(
            self,
            "parent_id",
            parent_id,
        )
        self.scope_id: str
        object.__setattr__(
            self,
            "scope_id",
            scope_id,
        )
        self.label: str
        object.__setattr__(
            self,
            "label",
            label,
        )
        self.unique_name: str
        object.__setattr__(
            self,
            "unique_name",
            f"[{trace_id}] [{label}] [{scope_id}]",
        )
        self._token: Token[ScopeIdentifier] | None
        object.__setattr__(
            self,
            "_token",
            None,
        )

    def __setattr__(
        self,
        name: str,
        value: Any,
    ) -> Any:
        raise AttributeError(
            f"Can't modify immutable {self.__class__.__qualname__},"
            f" attribute - '{name}' cannot be modified"
        )

    def __delattr__(
        self,
        name: str,
    ) -> None:
        raise AttributeError(
            f"Can't modify immutable {self.__class__.__qualname__},"
            f" attribute - '{name}' cannot be deleted"
        )

    @property
    def is_root(self) -> bool:
        return self.trace_id == self.parent_id

    def __str__(self) -> str:
        return self.unique_name

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, self.__class__):
            return False

        return self.scope_id == other.scope_id and self.trace_id == other.trace_id

    def __hash__(self) -> int:
        return hash(self.scope_id)

    def __enter__(self) -> None:
        assert self._token is None, "Context reentrance is not allowed"  # nosec: B101
        object.__setattr__(
            self,
            "_token",
            ScopeIdentifier._context.set(self),
        )

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        assert self._token is not None, "Unbalanced context enter/exit"  # nosec: B101
        ScopeIdentifier._context.reset(self._token)
        object.__setattr__(
            self,
            "_token",
            None,
        )



from haiway.types.default import Default, DefaultValue
from haiway.types.frozen import frozenlist
from haiway.types.missing import MISSING, Missing, is_missing, not_missing, when_missing

__all__ = [
    "MISSING",
    "Default",
    "DefaultValue",
    "Missing",
    "frozenlist",
    "is_missing",
    "not_missing",
    "when_missing",
]



from typing import Any, Final, TypeGuard, cast, final

__all__ = [
    "MISSING",
    "Missing",
    "is_missing",
    "not_missing",
    "when_missing",
]


class MissingType(type):
    _instance: Any = None

    def __call__(cls) -> Any:
        if cls._instance is None:
            cls._instance = super().__call__()
            return cls._instance

        else:
            return cls._instance


@final
class Missing(metaclass=MissingType):
    """
    Type representing absence of a value. Use MISSING constant for its value.
    """

    __slots__ = ()
    __match_args__ = ()

    def __bool__(self) -> bool:
        return False

    def __eq__(
        self,
        value: object,
    ) -> bool:
        return value is MISSING

    def __str__(self) -> str:
        return "MISSING"

    def __repr__(self) -> str:
        return "MISSING"

    def __getattr__(
        self,
        name: str,
    ) -> Any:
        raise AttributeError("Missing has no attributes")

    def __setattr__(
        self,
        __name: str,
        __value: Any,
    ) -> None:
        raise AttributeError("Missing can't be modified")

    def __delattr__(
        self,
        __name: str,
    ) -> None:
        raise AttributeError("Missing can't be modified")


MISSING: Final[Missing] = Missing()


def is_missing(
    check: Any | Missing,
    /,
) -> TypeGuard[Missing]:
    return check is MISSING


def not_missing[Value](
    check: Value | Missing,
    /,
) -> TypeGuard[Value]:
    return check is not MISSING


def when_missing[Value](
    check: Value | Missing,
    /,
    *,
    value: Value,
) -> Value:
    if check is MISSING:
        return value

    else:
        return cast(Value, check)



import typing
from collections.abc import Callable, Mapping
from types import EllipsisType, GenericAlias
from typing import (
    Any,
    ClassVar,
    Generic,
    Self,
    TypeVar,
    cast,
    dataclass_transform,
    final,
    overload,
)
from weakref import WeakValueDictionary

from haiway.state.attributes import AttributeAnnotation, attribute_annotations
from haiway.state.path import AttributePath
from haiway.state.validation import AttributeValidation, AttributeValidator
from haiway.types import MISSING, DefaultValue, Missing, not_missing

__all__ = [
    "State",
]


@overload
def Default[Value](
    value: Value,
    /,
) -> Value: ...


@overload
def Default[Value](
    *,
    factory: Callable[[], Value],
) -> Value: ...


def Default[Value](
    value: Value | Missing = MISSING,
    /,
    *,
    factory: Callable[[], Value] | Missing = MISSING,
) -> Value:  # it is actually a DefaultValue, but type checker has to be fooled
    return cast(Value, DefaultValue(value, factory=factory))


@final
class StateAttribute[Value]:
    __slots__ = (
        "annotation",
        "default",
        "name",
        "validator",
    )

    def __init__(
        self,
        name: str,
        annotation: AttributeAnnotation,
        default: DefaultValue[Value],
        validator: AttributeValidation[Value],
    ) -> None:
        self.name: str
        object.__setattr__(
            self,
            "name",
            name,
        )
        self.annotation: AttributeAnnotation
        object.__setattr__(
            self,
            "annotation",
            annotation,
        )
        self.default: DefaultValue[Value]
        object.__setattr__(
            self,
            "default",
            default,
        )
        self.validator: AttributeValidation[Value]
        object.__setattr__(
            self,
            "validator",
            validator,
        )

    def validated(
        self,
        value: Any | Missing,
        /,
    ) -> Value:
        return self.validator(self.default() if value is MISSING else value)

    def __setattr__(
        self,
        name: str,
        value: Any,
    ) -> Any:
        raise AttributeError(
            f"Can't modify immutable {self.__class__.__qualname__},"
            f" attribute - '{name}' cannot be modified"
        )

    def __delattr__(
        self,
        name: str,
    ) -> None:
        raise AttributeError(
            f"Can't modify immutable {self.__class__.__qualname__},"
            f" attribute - '{name}' cannot be deleted"
        )


@dataclass_transform(
    kw_only_default=True,
    frozen_default=True,
    field_specifiers=(DefaultValue,),
)
class StateMeta(type):
    def __new__(
        cls,
        /,
        name: str,
        bases: tuple[type, ...],
        namespace: dict[str, Any],
        type_parameters: dict[str, Any] | None = None,
        **kwargs: Any,
    ) -> Any:
        state_type = type.__new__(
            cls,
            name,
            bases,
            namespace,
            **kwargs,
        )

        attributes: dict[str, StateAttribute[Any]] = {}

        for key, annotation in attribute_annotations(
            state_type,
            type_parameters=type_parameters or {},
        ).items():
            default: Any = getattr(state_type, key, MISSING)
            attributes[key] = StateAttribute(
                name=key,
                annotation=annotation.update_required(default is MISSING),
                default=_resolve_default(default),
                validator=AttributeValidator.of(
                    annotation,
                    recursion_guard={
                        str(AttributeAnnotation(origin=state_type)): state_type.validator
                    },
                ),
            )

        state_type.__TYPE_PARAMETERS__ = type_parameters  # pyright: ignore[reportAttributeAccessIssue]
        state_type.__ATTRIBUTES__ = attributes  # pyright: ignore[reportAttributeAccessIssue]
        state_type.__slots__ = frozenset(attributes.keys())  # pyright: ignore[reportAttributeAccessIssue]
        state_type.__match_args__ = state_type.__slots__  # pyright: ignore[reportAttributeAccessIssue]
        state_type._ = AttributePath(state_type, attribute=state_type)  # pyright: ignore[reportCallIssue, reportUnknownMemberType, reportAttributeAccessIssue]

        return state_type

    def validator(
        cls,
        value: Any,
        /,
    ) -> Any: ...

    def __instancecheck__(
        self,
        instance: Any,
    ) -> bool:
        # check for type match
        if self.__subclasscheck__(type(instance)):  # pyright: ignore[reportUnknownArgumentType]
            return True

        # otherwise check if we are dealing with unparametrized base
        # against the parametrized one, our generic subtypes have base of unparametrized type
        if type(instance) not in self.__bases__:
            return False

        try:
            # validate instance to check unparametrized fields
            _ = self(**vars(instance))

        except Exception:
            return False

        else:
            return True

    def __subclasscheck__(  # noqa: C901, PLR0911, PLR0912
        self,
        subclass: type[Any],
    ) -> bool:
        # check if we are the same class for early exit
        if self == subclass:
            return True

        # then check if we are parametrized
        checked_parameters: Mapping[str, Any] | None = getattr(
            self,
            "__TYPE_PARAMETERS__",
            None,
        )
        if checked_parameters is None:
            # if we are not parametrized allow any subclass
            return self in subclass.__bases__

        # verify if we have common base next - our generic subtypes have the same base
        if self.__bases__ == subclass.__bases__:
            # if we have the same bases we have different generic subtypes
            # we can verify all of the attributes to check if we have common base
            available_parameters: Mapping[str, Any] | None = getattr(
                subclass,
                "__TYPE_PARAMETERS__",
                None,
            )

            if available_parameters is None:
                # if we have no parameters at this stage this is a serious bug
                raise RuntimeError("Invalid type parametrization for %s", subclass)

            for key, param in checked_parameters.items():
                match available_parameters.get(key):
                    case None:  # if any parameter is missing we should not be there already
                        return False

                    case typing.Any:
                        continue  # Any ignores type checks

                    case checked:
                        if param is Any:
                            continue  # Any ignores type checks

                        elif issubclass(checked, param):
                            continue  # if we have matching type we are fine

                        else:
                            return False  # types are not matching

            return True  # when all parameters were matching we have matching subclass

        elif subclass in self.__bases__:  # our generic subtypes have base of unparametrized type
            # if subclass parameters were not provided then we can be valid ony if all were Any
            return all(param is Any for param in checked_parameters.values())

        else:
            return False  # we have different base / comparing to not parametrized


def _resolve_default[Value](
    value: DefaultValue[Value] | Value | Missing,
) -> DefaultValue[Value]:
    if isinstance(value, DefaultValue):
        return cast(DefaultValue[Value], value)

    return DefaultValue[Value](
        value,
        factory=MISSING,
    )


_types_cache: WeakValueDictionary[
    tuple[
        Any,
        tuple[Any, ...],
    ],
    Any,
] = WeakValueDictionary()


class State(metaclass=StateMeta):
    """
    Base class for immutable data structures.
    """

    _: ClassVar[Self]
    __IMMUTABLE__: ClassVar[EllipsisType] = ...
    __TYPE_PARAMETERS__: ClassVar[Mapping[str, Any] | None] = None
    __ATTRIBUTES__: ClassVar[dict[str, StateAttribute[Any]]]

    @classmethod
    def __class_getitem__(
        cls,
        type_argument: tuple[type[Any], ...] | type[Any],
    ) -> type[Self]:
        assert Generic in cls.__bases__, "Can't specialize non generic type!"  # nosec: B101
        assert cls.__TYPE_PARAMETERS__ is None, "Can't specialize already specialized type!"  # nosec: B101

        type_arguments: tuple[type[Any], ...]
        match type_argument:
            case [*arguments]:
                type_arguments = tuple(arguments)

            case argument:
                type_arguments = (argument,)

        if any(isinstance(argument, TypeVar) for argument in type_arguments):  # pyright: ignore[reportUnnecessaryIsInstance]
            # if we got unfinished type treat it as an alias instead of resolving
            return cast(type[Self], GenericAlias(cls, type_arguments))

        assert len(type_arguments) == len(  # nosec: B101
            cls.__type_params__
        ), "Type arguments count has to match type parameters count"

        if cached := _types_cache.get((cls, type_arguments)):
            return cached

        type_parameters: dict[str, Any] = {
            parameter.__name__: argument
            for (parameter, argument) in zip(
                cls.__type_params__ or (),
                type_arguments or (),
                strict=False,
            )
        }

        parameter_names: str = ",".join(
            getattr(
                argument,
                "__name__",
                str(argument),
            )
            for argument in type_arguments
        )
        name: str = f"{cls.__name__}[{parameter_names}]"
        bases: tuple[type[Self]] = (cls,)

        parametrized_type: type[Self] = StateMeta.__new__(
            cls.__class__,
            name=name,
            bases=bases,
            namespace={"__module__": cls.__module__},
            type_parameters=type_parameters,
        )
        _types_cache[(cls, type_arguments)] = parametrized_type
        return parametrized_type

    @classmethod
    def validator(
        cls,
        value: Any,
        /,
    ) -> Self:
        match value:
            case validated if isinstance(validated, cls):
                return validated

            case {**values}:
                return cls(**values)

            case _:
                raise TypeError(f"Expected '{cls.__name__}', received '{type(value).__name__}'")

    def __init__(
        self,
        **kwargs: Any,
    ) -> None:
        for name, attribute in self.__ATTRIBUTES__.items():
            object.__setattr__(
                self,  # pyright: ignore[reportUnknownArgumentType]
                name,
                attribute.validated(
                    kwargs.get(
                        name,
                        MISSING,
                    ),
                ),
            )

    def updating[Value](
        self,
        path: AttributePath[Self, Value] | Value,
        /,
        value: Value,
    ) -> Self:
        assert isinstance(  # nosec: B101
            path, AttributePath
        ), "Prepare parameter path by using Self._.path.to.property or explicitly"

        return cast(AttributePath[Self, Value], path)(self, updated=value)

    def updated(
        self,
        **kwargs: Any,
    ) -> Self:
        return self.__replace__(**kwargs)

    def as_dict(self) -> dict[str, Any]:
        dict_result: dict[str, Any] = {}
        for key in self.__ATTRIBUTES__.keys():
            value: Any | Missing = getattr(self, key, MISSING)
            if not_missing(value):
                dict_result[key] = value

        return dict_result

    def __str__(self) -> str:
        attributes: str = ", ".join([f"{key}: {value}" for key, value in vars(self).items()])
        return f"{self.__class__.__name__}({attributes})"

    def __repr__(self) -> str:
        return str(self)

    def __eq__(
        self,
        other: Any,
    ) -> bool:
        if not issubclass(other.__class__, self.__class__):
            return False

        return all(
            getattr(self, key, MISSING) == getattr(other, key, MISSING)
            for key in self.__ATTRIBUTES__.keys()
        )

    def __setattr__(
        self,
        name: str,
        value: Any,
    ) -> Any:
        raise AttributeError(
            f"Can't modify immutable state {self.__class__.__qualname__},"
            f" attribute - '{name}' cannot be modified"
        )

    def __delattr__(
        self,
        name: str,
    ) -> None:
        raise AttributeError(
            f"Can't modify immutable state {self.__class__.__qualname__},"
            f" attribute - '{name}' cannot be deleted"
        )

    def __copy__(self) -> Self:
        return self  # State is immutable, no need to provide an actual copy

    def __deepcopy__(
        self,
        memo: dict[int, Any] | None,
    ) -> Self:
        return self  # State is immutable, no need to provide an actual copy

    def __replace__(
        self,
        **kwargs: Any,
    ) -> Self:
        return self.__class__(
            **{
                **vars(self),
                **kwargs,
            }
        )
